import torch
import torch.nn as nn
from torch.nn import TransformerEncoder, TransformerEncoderLayer
from src.utils import get_device


class MeanMaxPooling(nn.Module):
    """
    (C, B, E) -> (B, 2*E)
    where C is context count, B is batch size and E is embedding size.
    """

    def __init__(self, axis=0, dropout=0.0):
        super(MeanMaxPooling, self).__init__()
        self.axis = axis
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, inputs, valid_length=None):
        mean_out = torch.mean(inputs, dim=self.axis) if valid_length is None \
            else torch.sum(inputs, dim=self.axis) / valid_length
        max_out = torch.max(inputs, dim=self.axis).values
        outputs = torch.cat((mean_out, max_out), dim=1)
        outputs = self.dropout(outputs)
        return outputs


class ContextTransformer(nn.Module):
    """
    Transform context value to encoded embeddings
    [(B, 1) * C] -> (B, cross_size)
    """

    def __init__(
            self,
            context_dims,
            context_embed_size,
            context_num_heads,
            context_hidden_size,
            context_transformer_dropout,
            context_num_layers,
            context_pooling_dropout,
            cross_size,
            device,
    ):
        super(ContextTransformer, self).__init__()
        self.context_embedding = [nn.Embedding(context_dim, context_embed_size).to(device)
                                  for context_dim in context_dims]
        encoder_layers = TransformerEncoderLayer(d_model=context_embed_size,
                                                 nhead=context_num_heads,
                                                 dropout=context_transformer_dropout,
                                                 dim_feedforward=context_hidden_size,
                                                 activation='relu').to(device)
        self.context_encoder = TransformerEncoder(encoder_layers, num_layers=context_num_layers).to(device)
        self.context_pooling_dp = MeanMaxPooling(dropout=context_pooling_dropout)
        self.context_dense = torch.nn.Linear(in_features=2 * context_embed_size, out_features=cross_size).to(device)

    def forward(self, input_context_list):
        """
        context_input should be a list of tensors like [(B, 1) * S]
        """
        # [(B, 1) * C]
        context_embedding_list = [self.context_embedding[i](input_context)
                                  for i, input_context in enumerate(input_context_list)]  # -> [(B, 1, E) * C]
        context_out = torch.cat(context_embedding_list, dim=1).transpose(0, 1)  # -> (B, C, E) -> (C, B, E)

        context_out = self.context_encoder(context_out)  # -> (C, B, E)
        context_out = self.context_pooling_dp(context_out)  # -> (B, 2*E)
        context_out = self.context_dense(context_out)  # -> (B, cross_size)
        # print(context_out.shape)
        return context_out


class SequenceTransformer(nn.Module):
    """
    (B, 5) -> (B, cross_size)
    """

    def __init__(
            self,
            item_dim,
            item_embed_size,
            item_max_length,
            item_num_heads,
            item_hidden_size,
            item_transformer_dropout,
            item_num_layers,
            item_pooling_dropout,
            item_embedding_weight,
            cross_size,
            device,
    ):
        super(SequenceTransformer, self).__init__()
        if item_embedding_weight is not None:
            self.item_embedding = nn.Embedding.from_pretrained(item_embedding_weight, freeze=True).to(device)
        else:
            self.item_embedding = nn.Embedding(item_dim, item_embed_size).to(device)
        encoder_layers = TransformerEncoderLayer(d_model=item_embed_size,
                                                 nhead=item_num_heads,
                                                 dropout=item_transformer_dropout,
                                                 dim_feedforward=item_hidden_size,
                                                 activation='relu').to(device)
        self.item_encoder = TransformerEncoder(encoder_layers, num_layers=item_num_layers).to(device)
        self.item_pooling_dp = MeanMaxPooling(dropout=item_pooling_dropout)
        self.item_dense = torch.nn.Linear(in_features=2 * item_embed_size, out_features=cross_size).to(device)

    def forward(self, input_item_seq):
        # print("initial_shape:",input_item.shape)
        # (B, 5)
        item_embed_out = self.item_embedding(input_item_seq.long())  # -> (B, 5, E)
        item_embed_out = item_embed_out.transpose(0, 1)  # -> (B, 5, E) -> (5, B, E)
        item_out = self.item_encoder(item_embed_out)
        item_out = self.item_pooling_dp(item_out)  # -> (B, 2*E)
        item_out = self.item_dense(item_out)  # -> (B, cross_size)

        return item_out


class TxTFeaturesExtractor(nn.Module):
    """

    """

    def __init__(
            self,
            item_dim,
            context_dims,
            act_type="gelu",
            cross_size=100,
            context_embed_size=100,
            context_num_heads=2,
            context_hidden_size=512,
            context_transformer_dropout=0.0,
            context_num_layers=2,
            context_pooling_dropout=0.0,
            item_max_length=8,  # TODO: not using now
            item_embed_size=100,
            item_num_heads=4,
            item_hidden_size=256,
            item_transformer_dropout=0.0,
            item_num_layers=2,
            item_pooling_dropout=0.1,
            item_embedding_weight=None,
            device="auto",
    ):
        super(TxTFeaturesExtractor, self).__init__()
        device = get_device(device)
        self.features_dim = cross_size
        self.context_transformer = ContextTransformer(
            context_dims=context_dims,
            context_embed_size=context_embed_size,
            context_num_heads=context_num_heads,
            context_hidden_size=context_hidden_size,
            context_transformer_dropout=context_transformer_dropout,
            context_num_layers=context_num_layers,
            context_pooling_dropout=context_pooling_dropout,
            cross_size=cross_size,
            device=device,
        )
        self.sequence_transformer = SequenceTransformer(
            item_dim=item_dim,
            item_embed_size=item_embed_size,
            item_max_length=item_max_length,
            item_num_heads=item_num_heads,
            item_hidden_size=item_hidden_size,
            item_transformer_dropout=item_transformer_dropout,
            item_num_layers=item_num_layers,
            item_pooling_dropout=item_pooling_dropout,
            item_embedding_weight=item_embedding_weight,
            cross_size=cross_size,
            device=device,
        )
        self.dense1 = nn.Linear(cross_size, item_dim // 2).to(device)
        if act_type == "relu":
            self.act = nn.ReLU()
        elif act_type == "gelu":
            self.act = nn.GELU()
        elif act_type == "leakyRelu":
            self.act = nn.LeakyReLU(0.2)
        else:
            raise NotImplementedError
        self.dense2 = nn.Linear(item_dim // 2, item_dim).to(device)

    def forward(self, input_context_list, input_item_seq):
        # input [(B, 1) * C] and (B, 5)
        context_outs = self.context_transformer(input_context_list)  # -> (B, cross_size)
        item_outs = self.sequence_transformer(input_item_seq)  # -> (B.cross_size)
        outs = torch.mul(item_outs, context_outs)  # -> (B, cross_size)
        outs = self.dense1(outs)  # -> (B, item_dim//2)
        outs = self.act(outs)
        outs = self.dense2(outs)  # -> (B, item_dim)
        return outs



import abc
import wandb
import pytorch_lightning as pl


class MLflowCallbackBase(pl.callbacks.Callback):
    """
    This callback updates metrics on MLflow after every train epoch and validation epoch

    Abstract Method::
        - update_mlflow_metrics
    """

    def on_validation_epoch_start(self, trainer, pl_module):
        if not trainer.sanity_checking:
            self.update_mlflow_metrics(trainer, pl_module, "Train")

    def on_validation_epoch_end(self, trainer, pl_module):
        self.update_mlflow_metrics(trainer, pl_module, "Validation")

    @abc.abstractmethod
    def update_mlflow_metrics(self, trainer, pl_module, phase):
        """
        Log metrics to mlflow using mlflow.log_metrics

        :param trainer: trainer passed from on_*_epoch_end methods, you can find your logger, callback_metrics and logged_metrics here
        :param pl_module: the trained model passed from on_*_epoch_end methods
        :param phase: Train or Validation phase in the training process

        Example:

        def update_mlflow_metrics(self, trainer, pl_module, phase):
            mlflow.log_metrics(
                metrics={f"{phase}/loss": loss_from_training,
                         f"{phase}/acc": acc_from_trainer},
                step=pl_module.current_epoch
            )
        """
        pass


class TensorboardCallbackBase(pl.callbacks.Callback):
    """
    Update model parameter histograms and metric scalars on tensorboard

    Abstract Method::
        - update_metrics_scalar

    Attributes::
        - tensorboard_logger: TensorBoardLogger from pytorch lightning to add graphs
    """

    def __init__(self, tensorboard_logger):
        """
        :param tensorboard_logger: TensorBoardLogger from pytorch lightning to add graphs
        """
        self.tensorboard_logger = tensorboard_logger

    def on_validation_epoch_start(self, trainer, pl_module):
        if not trainer.sanity_checking:
            self.update_parameter_histogram(pl_module)
            self.update_metrics_scalar(trainer, pl_module, "Train")

    def on_validation_epoch_end(self, trainer, pl_module):
        self.update_metrics_scalar(trainer, pl_module, "Validation")

    def update_parameter_histogram(self, pl_module):
        for name, params in pl_module.named_parameters():
            self.tensorboard_logger.experiment.add_histogram(name, params, pl_module.current_epoch, max_bins=512)

    @abc.abstractmethod
    def update_metrics_scalar(self, trainer, pl_module, phase):
        """
        update scalar metrics to tensorboard

        :param trainer: trainer passed from on_*_epoch_end methods, you can find your logger, callback_metrics and logged_metrics here
        :param pl_module: the trained model passed from on_*_epoch_end methods
        :param phase: Train or Validation phase in the training process

        Example::

        def update_metrics_scalar(self, trainer, pl_module, phase):
            epoch = pl_module.current_epoch
            trainer_logger = trainer.logger
            self.tensorboard_logger.experiment.add_scalar(f"loss/{phase}", loss_from_trainer_logger, epoch)
            self.tensorboard_logger.experiment.add_scalar(f"acc/{phase}", acc_from_trainer_logger, epoch)
        """
        pass


class AttnVisCallbackBase(pl.callbacks.Callback):
    """
    Visualize attention importance using wandb's line_series plot

    Abstract Method::
        - get_attn_weights

    Attributes::
        - transformer_prefix: a str indicating what type of transformer is this e.g. context/sequence
        - cols: input feature columns
        - sample_inputs: sample inputs to generate attention weights
        - epochs: list of epoch numbers
        - avg_attn_weights: list of lists containing average attention weights for each epoch
        - device: gpu or cpu device
    """

    def __init__(self, transformer_prefix, cols, sample_inputs, device):
        """
        :param transformer_prefix: a str indicating what type of transformer is this e.g. context/sequence
        :param cols: input feature columns
        :param sample_inputs: sample inputs to generate attention weights
        """
        self.transformer_prefix = transformer_prefix
        self.cols = cols
        self.sample_inputs = sample_inputs
        self.epochs = []
        self.avg_attn_weights = [[] for _ in range(len(self.cols))]
        self.device = device

    def on_validation_epoch_end(self, trainer, pl_module):
        self.epochs.append(pl_module.current_epoch)
        avg_attn_weight = self.get_attn_weights(pl_module)
        for i in range(len(self.cols)):
            self.avg_attn_weights[i].append(avg_attn_weight[i])
        wandb.log(
            {
                f"attention_importance_line_{self.transformer_prefix}": wandb.plot.line_series(
                    xs=self.epochs, ys=self.avg_attn_weights,
                    keys=self.cols,
                    title=f"Attention Importance Line - {self.transformer_prefix}",
                    xname="step"
                )
            }
        )

    @abc.abstractmethod
    def get_attn_weights(self, pl_module):
        """
        Using self.sample_inputs, calculate average attention on input features

        :param pl_module: the trained model passed from on_validation_epoch_end method
        :return: a list of averaged attention in the order of self.cols

        Example::

        def get_attn_weights(self, pl_module):
            tmp_in = self.sample_inputs.to(self.device)
            transformer_layer = pl_module.transformer_layer
            attn_layer = transformer_layer.encoder.layers[0].self_attn
            tmp_out = transformer_layer.embedding_layer(tmp_in)
            _, attn_weight = attn_layer(tmp_out, tmp_out, tmp_out, need_weights=True)
            avg_attn_weight = attn_weight.mean(0).mean(0).detach().tolist()
            return avg_attn_weight
        """
        return []



import os
import logging
from datetime import datetime
import pandas as pd
import boto3


def time2weekday(time):
    """
    0 -> 6
    Mon -> Sun
    """
    return int(datetime.fromisoformat(time).weekday())


def time2daytime(time):
    dt = datetime.fromisoformat(time)
    h = int(dt.hour)
    m = int(dt.minute)
    return 2 * h if m < 30 else 2 * h + 1


def lookup_table_to_dict(table_path, idx_col, val_col, file_format="json"):
    if file_format == "json":
        lookup = pd.read_json(table_path, lines=True)
    elif file_format == "csv":
        lookup = pd.read_csv(table_path)
    else:
        raise NotImplementedError()
    lookup_dict = lookup[[idx_col, val_col]].set_index(idx_col).to_dict()[val_col]
    return lookup_dict


class DownloadS3(object):
    def __init__(self, bucket, access_key, access_secret):
        logging.basicConfig(
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            level=logging.INFO,
        )
        self.logger = logging.getLogger("app_download")
        self.bucket = bucket
        self.access_key = access_key
        self.access_secret = access_secret
        self.data = None

    def download_folder(self, s3_prefix, local_dir):
        s3 = boto3.client('s3',
                          aws_access_key_id=self.access_key,
                          aws_secret_access_key=self.access_secret,
                          verify=False)
        list_obj = s3.list_objects_v2(Bucket=self.bucket, Prefix=s3_prefix)
        if list_obj["KeyCount"] > 1000:
            self.logger.error("Found more than 1000 keys and only read partial data")
        for content in list_obj["Contents"]:
            key = content["Key"]
            if not key.endswith("/"):
                path, filename = os.path.split(key)
                local_filepath = os.path.join(local_dir, filename)

                self.logger.info(key)
                self.logger.info(local_filepath)
                s3.download_file(self.bucket, key, local_filepath)
                self.logger.info("Download s3 {bucket}/{s3_path} to local {local_path}".format(bucket=self.bucket,
                                                                                          s3_path=key,
                                                                                          local_path=local_filepath))
        return None

    def download_single(self, s3_prefix, local_path, file_format="csv"):
        """
        Concat and download Databricks generated csv/json folder to local csv and rename it.
        """
        dfs = []
        keys = []
        s3 = boto3.client('s3',
                          aws_access_key_id=self.access_key,
                          aws_secret_access_key=self.access_secret,
                          verify=False)
        list_obj = s3.list_objects_v2(Bucket=self.bucket, Prefix=s3_prefix)
        if list_obj["KeyCount"] > 1000:
            self.logger.error("Found more than 1000 keys and only read partial data")
        for content in list_obj["Contents"]:
            if content["Key"].endswith(file_format):
                keys.append(content["Key"])
        self.logger.info("Found {n} {format} files from {bucket}/{dir}".format(n=len(keys), format=file_format,
                                                                          bucket=self.bucket, dir=s3_prefix))

        for key in keys:
            obj = s3.get_object(Bucket=self.bucket, Key=key)
            if file_format == "json":
                df = pd.read_json(obj["Body"], orient="columns", lines=True)
            elif file_format == "csv":
                df = pd.read_csv(obj["Body"])
            else:
                raise NotImplementedError()
            dfs.append(df)
        df = pd.concat(dfs)
        df.to_csv(local_path, index=False)
        self.logger.info("Save to {}".format(local_path))
        return None


import math
import torch
import torch.nn as nn
from torch.nn import TransformerEncoder, TransformerEncoderLayer


class MeanMaxPooling(nn.Module):
    """
    [B, S, E] -> [B, 2*E]
    """
    def __init__(self, axis=1, dropout=0.0):
        super().__init__()
        self.axis = axis
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, inputs, valid_length=None):
        """
        :param inputs: Tensor, shape [batch_size, seq_len, embedding_dim]
        :param valid_length: None or Tensor, valid len of token in the sequence with shape [batch_size]
        :return: Tensor, shape [batch_size, 2 * embedding_dim]
        """
        # TODO: broadcast indexing to mean over first vl
        mean_out = torch.mean(inputs, dim=self.axis) if valid_length is None \
            else torch.sum(inputs, dim=self.axis) / valid_length.add(1E-7).unsqueeze(1)
        max_out = torch.max(inputs, dim=self.axis).values
        outputs = torch.cat((mean_out, max_out), dim=1)
        outputs = self.dropout(outputs)
        return outputs


class PositionalEncoding(nn.Module):
    """
    [B, S, E] -> [B, S, E]
    """
    def __init__(self, d_model, max_len=5000, dropout=0.1):
        """
        Constant layer of positional encoding

        :param d_model: int, dimension of token embedding
        :param max_len: int, max length of sequence
        :param dropout: float
        """
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x):
        """
        :param x: Tensor, shape [batch_size, seq_len, embedding_dim]
        :return: : Tensor, shape [batch_size, seq_len, embedding_dim]
        """
        x = x + self.pe[:, :x.size(1), :]
        return self.dropout(x)


class ContextTransformer(nn.Module):
    """
    [[B, ] * C] -> [B, cross_size]
    """
    def __init__(self, ctx_nums, cross_size, ctx_embed_size=100, ctx_num_heads=2,
                 ctx_hidden_size=512, ctx_num_layers=1, ctx_transformer_dropout=0.0,
                 ctx_pooling_dropout=0.0, ctx_pe=False):
        super().__init__()
        self.ctx_embedding = nn.ModuleList([
            nn.Embedding(ctx_num, ctx_embed_size)
            for ctx_num in ctx_nums
        ])
        self.ctx_pe = ctx_pe
        self.ctx_embed_size = ctx_embed_size
        if ctx_pe:
            self.pos_encoder = PositionalEncoding(d_model=ctx_embed_size,
                                                  dropout=ctx_transformer_dropout,
                                                  max_len=len(ctx_nums))
        encoder_layers = TransformerEncoderLayer(d_model=ctx_embed_size,
                                                 nhead=ctx_num_heads,
                                                 dropout=ctx_transformer_dropout,
                                                 dim_feedforward=ctx_hidden_size,
                                                 activation='relu',
                                                 batch_first=True)
        self.ctx_encoder = TransformerEncoder(encoder_layers, num_layers=ctx_num_layers)
        self.ctx_pooling_dp = MeanMaxPooling(dropout=ctx_pooling_dropout)
        self.ctx_dense = torch.nn.Linear(in_features=2 * ctx_embed_size, out_features=cross_size)

    def forward(self, ctx_in):
        """
        :param ctx_in: list, a list of Tensor of shape [batch_size, 1]
        :return: Tensor, shape [batch_size, cross_size]
        """
        # [[B, ] * C]
        ctx_embedding_list = [self.ctx_embedding[i](input_ctx).unsqueeze(1)
                                  for i, input_ctx in enumerate(ctx_in)]  # -> [(B, 1, E) * C]
        ctx_out = torch.cat(ctx_embedding_list, dim=1)  # -> (B, C, E)
        if self.ctx_pe:
            ctx_out = ctx_out * math.sqrt(self.ctx_embed_size)
            ctx_out = self.pos_encoder(ctx_out)
        ctx_out = self.ctx_encoder(ctx_out)  # -> (B, C, E)
        ctx_out = self.ctx_pooling_dp(ctx_out)  # -> (B, 2*E)
        ctx_out = self.ctx_dense(ctx_out)  # -> (B, cross_size)
        # print(ctx_out.shape)
        return ctx_out


class SequenceTransformerHistory(nn.Module):
    """
    [B, S] -> [B, cross_size]
    """

    def __init__(self, seq_num, cross_size, seq_embed_size=200, seq_max_length=8, seq_num_heads=4,
                 seq_hidden_size=512, seq_transformer_dropout=0.0, seq_num_layers=2, seq_pooling_dropout=0.0,
                 seq_pe=True):
        super().__init__()
        self.seq_embedding = nn.Embedding(seq_num, seq_embed_size)
        self.seq_pos = seq_pe
        self.seq_embed_size = seq_embed_size
        if seq_pe:
            self.pos_encoder = PositionalEncoding(d_model=seq_embed_size,
                                                  dropout=seq_transformer_dropout,
                                                  max_len=seq_max_length)
        encoder_layers = TransformerEncoderLayer(d_model=seq_embed_size,
                                                 nhead=seq_num_heads,
                                                 dropout=seq_transformer_dropout,
                                                 dim_feedforward=seq_hidden_size,
                                                 activation='relu',
                                                 batch_first=True)
        self.seq_encoder = TransformerEncoder(encoder_layers, num_layers=seq_num_layers)
        self.seq_pooling_dp = MeanMaxPooling(dropout=seq_pooling_dropout)
        self.seq_dense = torch.nn.Linear(in_features=2 * seq_embed_size, out_features=cross_size)

    @staticmethod
    def create_key_padding_mask(seq_in, valid_length):
        device = valid_length.device
        # mask = torch.arange(seq_in.size(1)).repeat(seq_in.size(0), 1).to(device)
        mask = torch.arange(seq_in.size(1)).repeat(seq_in.size(0), 1).to(device)
        mask = ~mask.lt(valid_length.unsqueeze(1))
        return mask

    def forward(self, seq_in, vl_in, seq_history=None):
        """
        :param seq_in: Tensor, shape [batch_size, seq_len]
        :param vl_in: Tensor, shape [batch_size]
        :param seq_history: Tensor, shape [batch_size, history_len]
        :return: Tensor, shape [batch_size, cross_size]
        """
        # print("initial_shape:",input_seq.shape)
        # (B, 5), (B, 10)
        seq_embed_out = self.seq_embedding(seq_in.long())  # -> (B, 5, E)
        # history_embed_out = self.seq_embedding(input_history_seq.long())
        # history_embed_out = history_embed_out.transpose(0, 1).mean(dim=0, keepdim=True)  # -> (1, B, E)
        # combined_embed_out = torch.cat([history_embed_out, seq_embed_out], dim=0)  # -> (6, B, E)
        seq_out = seq_embed_out
        if self.seq_pos:
            seq_out = seq_out * math.sqrt(self.seq_embed_size)
            seq_out = self.pos_encoder(seq_out)
        mask = self.create_key_padding_mask(seq_in=seq_in, valid_length=vl_in)
        seq_out = self.seq_encoder(seq_out, src_key_padding_mask=mask)
        if mask[:, 0].any():
            seq_out = seq_out.nan_to_num(nan=0.0)
        seq_out = self.seq_pooling_dp(seq_out)  # -> (B, 2*E)
        seq_out = self.seq_dense(seq_out)  # -> (B, cross_size)
        return seq_out


class TxT(nn.Module):
    def __init__(self, ctx_nums, seq_num, act_type="gelu", cross_size=200,
                 context_transformer_kwargs=None, sequence_transformer_kwargs=None):
        super().__init__()
        context_transformer_kwargs = context_transformer_kwargs if context_transformer_kwargs else {}
        sequence_transformer_kwargs = sequence_transformer_kwargs if sequence_transformer_kwargs else {}
        self.features_dim = cross_size
        self.context_transformer = ContextTransformer(
            ctx_nums=ctx_nums,
            cross_size=cross_size,
            **context_transformer_kwargs,
        )
        self.sequence_transformer = SequenceTransformerHistory(
            seq_num=seq_num,
            cross_size=cross_size,
            **sequence_transformer_kwargs,
        )
        self.dense1 = nn.Linear(cross_size, seq_num // 2)
        if act_type == "relu":
            self.act = nn.ReLU()
        elif act_type == "gelu":
            self.act = nn.GELU()
        elif act_type == "leakyRelu":
            self.act = nn.LeakyReLU(0.2)
        else:
            raise NotImplementedError
        self.dense2 = nn.Linear(seq_num // 2, seq_num)

    def forward(self, ctx_in, seq_in, vl_in, seq_history=None):
        """
        :param ctx_in: list, a list of Tensor of shape [batch_size]
        :param seq_in: Tensor, shape [batch_size, seq_len]
        :param vl_in: Tensor, shape [batch_size]
        :param seq_history: Tensor, shape [batch_size, history_len]
        :return:
        """
        # input [[B, ] * C] and [B, 5]
        ctx_out = self.context_transformer(ctx_in=ctx_in)
        seq_out = self.sequence_transformer(seq_in=seq_in, vl_in=vl_in, seq_history=seq_history)
        outs = torch.mul(seq_out, ctx_out)  # -> [B, cross_size]
        outs = self.dense1(outs)  # -> [B, seq_num//2]
        outs = self.act(outs)
        outs = self.dense2(outs)  # -> [B, seq_num]
        return outs



import time
import math
import logging
import itertools
import functools
from contextlib import contextmanager
import numpy as np
import gluonnlp as nlp
import mxnet as mx


logger = logging.getLogger("offer_encoder_offer2vec")


# Help functions - utils

def get_context(args):
    if args.gpu is None or args.gpu == '':
        context = [mx.cpu()]
    elif isinstance(args.gpu, int):
        context = [mx.gpu(args.gpu)]
    else:
        context = [mx.gpu(int(i)) for i in args.gpu]
    return context


@contextmanager
def print_time(task):
    start_time = time.time()
    logging.info('Starting to %s', task)
    yield
    logging.info('Finished to {} in {:.2f} seconds'.format(
        task,
        time.time() - start_time))


# Helper functions - data

def preprocess_dataset(data, vocab_len=None, min_freq=1, max_vocab_size=None):
    with print_time('count and construct vocabulary'):
        counter = nlp.data.count_tokens(itertools.chain.from_iterable(data))
        if vocab_len:
            counter.update(range(vocab_len))
            logger.info("Build vocab with indexed data of length %d" % vocab_len)
            vocab = nlp.Vocab(
                counter,
                token_to_idx={x: x for x in counter.keys()},
                unknown_token=None,
                mask_token=None,
                padding_token=0,
                bos_token=None,
                eos_token=None,
                min_freq=min_freq,
                max_size=max_vocab_size
            )
        else:
            logger.info("Build vocab with unindexed data")
            vocab = nlp.Vocab(
                counter,
                unknown_token=None,
                mask_token=None,
                padding_token=0,
                bos_token=None,
                eos_token=None,
                min_freq=min_freq,
                max_size=max_vocab_size
            )
        idx_to_counts = [counter[w] for w in vocab.idx_to_token]

    def code(sentence):
        return [vocab[token] for token in sentence if token in vocab]

    with print_time('code data'):
        data = data.transform(code, lazy=False)
    data = nlp.data.SimpleDataStream([data])
    return data, vocab, idx_to_counts


def transform_data_word2vec(data, vocab, idx_to_counts, cbow, batch_size,
                            window_size, frequent_token_subsampling=1E-4,
                            dtype='float32', index_dtype='int64'):
    """Transform a DataStream of coded DataSets to a DataStream of batches.

    Parameters
    ----------
    data : gluonnlp.data.DataStream
        DataStream where each sample is a valid input to
        gluonnlp.data.EmbeddingCenterContextBatchify.
    vocab : gluonnlp.Vocab
        Vocabulary containing all tokens whose indices occur in data.
    idx_to_counts : list of int
        List of integers such that idx_to_counts[idx] represents the count of
        vocab.idx_to_token[idx] in the underlying dataset. The count
        information is used to subsample frequent words in the dataset.
        Each token is independently dropped with probability 1 - sqrt(t /
        (count / sum_counts)) where t is the hyperparameter
        frequent_token_subsampling.
    cbow : boolean
        If True, use CBOW. Otherwise, use SkipGram
    batch_size : int
        The returned data stream iterates over batches of batch_size.
    window_size : int
        The context window size for
        gluonnlp.data.EmbeddingCenterContextBatchify.
    frequent_token_subsampling : float
        Hyperparameter for subsampling. See idx_to_counts above for more
        information.
    dtype : str or np.dtype, default 'float32'
        Data type of data array.
    index_dtype : str or np.dtype, default 'int64'
        Data type of index arrays.

    Returns
    -------
    gluonnlp.data.DataStream
        Stream over batches.
    """

    sum_counts = float(sum(idx_to_counts))
    idx_to_pdiscard = [
        1 - math.sqrt(frequent_token_subsampling / (count / sum_counts))
        for count in idx_to_counts]

    def subsample(shard):
        return [[
            t for t, r in zip(sentence,
                              np.random.uniform(0, 1, size=len(sentence)))
            if r > idx_to_pdiscard[t]] for sentence in shard]

    data = data.transform(subsample)

    batchify = nlp.data.batchify.EmbeddingCenterContextBatchify(
        batch_size=batch_size, window_size=window_size, cbow=cbow,
        weight_dtype=dtype, index_dtype=index_dtype, shuffle=True)
    data = data.transform(batchify)
    data = UnchainStream(data)

    if cbow:
        batchify_fn = cbow_batch
    else:
        batchify_fn = skipgram_batch
    batchify_fn = functools.partial(batchify_fn, num_tokens=len(vocab),
                                    dtype=dtype, index_dtype=index_dtype)

    return data, batchify_fn,


class UnchainStream(nlp.data.DataStream):
    def __init__(self, iterable):
        self._stream = iterable

    def __iter__(self):
        return iter(itertools.chain.from_iterable(self._stream))


def cbow_batch(centers, contexts, num_tokens, dtype, index_dtype):
    """Create a batch for CBOW training objective."""
    contexts_data, contexts_row, contexts_col = contexts
    centers = mx.nd.array(centers, dtype=index_dtype)
    contexts = mx.nd.sparse.csr_matrix(
        (contexts_data, (contexts_row, contexts_col)),
        dtype=dtype, shape=(len(centers), num_tokens))  # yapf: disable
    return centers, contexts


def skipgram_batch(centers, contexts, num_tokens, dtype, index_dtype):
    """Create a batch for SG training objective."""
    contexts = mx.nd.array(contexts[2], dtype=index_dtype)
    indptr = mx.nd.arange(len(centers) + 1)
    centers = mx.nd.array(centers, dtype=index_dtype)
    centers_csr = mx.nd.sparse.csr_matrix(
        (mx.nd.ones(centers.shape), centers, indptr), dtype=dtype,
        shape=(len(centers), num_tokens))
    return centers_csr, contexts, centers


# Helper functions - model

class Net(mx.gluon.HybridBlock):
    """Base class for word2vec and fastText SkipGram and CBOW networks.

    Parameters
    ----------
    token_to_idx : dict
        token_to_idx mapping of the vocabulary that this model is to be trained
        with. token_to_idx is used for __getitem__ and __contains__. For
        len(token_to_idx) is used during initialization to obtain the input_dim
        of the embedding matrix.
    output_dim : int
        Dimension of the dense embedding.
    batch_size : int
        Batchsize this model will be trained with. TODO temporary until
        random_like ops are supported
    negatives_weights : mxnet.nd.NDArray
        Weights for UnigramCandidateSampler for sampling negatives.
    smoothing : float, default 0.75
        Smoothing factor applied to negatives_weights. Final weights are
        mxnet.nd.power(negative_weights, smoothing).
    num_negatives : int, default 5
        Number of negatives to sample for each real sample.
    sparse_grad : bool, default True
        Specifies mxnet.gluon.nn.Embedding sparse_grad argument.
    dtype : str, default 'float32'
        dtype argument passed to gluon.nn.Embedding

    """

    # pylint: disable=abstract-method
    def __init__(self, token_to_idx, output_dim, negatives_weights,
                 subword_function=None, num_negatives=5, smoothing=0.75,
                 sparse_grad=True, dtype='float32', **kwargs):
        super(Net, self).__init__(**kwargs)

        self._kwargs = dict(
            input_dim=len(token_to_idx), output_dim=output_dim, dtype=dtype,
            sparse_grad=sparse_grad, num_negatives=num_negatives)

        with self.name_scope():
            if subword_function is not None:
                self.embedding = nlp.model.train.FasttextEmbeddingModel(
                    token_to_idx=token_to_idx,
                    subword_function=subword_function,
                    output_dim=output_dim,
                    weight_initializer=mx.init.Uniform(scale=1 / output_dim),
                    sparse_grad=sparse_grad,
                )
            else:
                self.embedding = nlp.model.train.CSREmbeddingModel(
                    token_to_idx=token_to_idx,
                    output_dim=output_dim,
                    weight_initializer=mx.init.Uniform(scale=1 / output_dim),
                    sparse_grad=sparse_grad,
                )
            self.embedding_out = mx.gluon.nn.Embedding(
                len(token_to_idx), output_dim=output_dim,
                weight_initializer=mx.init.Zero(), sparse_grad=sparse_grad,
                dtype=dtype)

            self.negatives_sampler = nlp.data.UnigramCandidateSampler(
                weights=negatives_weights**smoothing, dtype='int64')

    def __getitem__(self, tokens):
        return self.embedding[tokens]

    def hybrid_forward(self, F, x, *args, **kwargs):
        raise NotImplementedError()


class SG(Net):
    """SkipGram network"""

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, center, context, center_words):
        """SkipGram forward pass.

        Parameters
        ----------
        center : mxnet.nd.NDArray or mxnet.sym.Symbol
            Sparse CSR array of word / subword indices of shape (batch_size,
            len(token_to_idx) + num_subwords). Embedding for center words are
            computed via F.sparse.dot between the CSR center array and the
            weight matrix.
        context : mxnet.nd.NDArray or mxnet.sym.Symbol
            Dense array of context words of shape (batch_size, ). Also used for
            row-wise independently masking negatives equal to one of context.
        center_words : mxnet.nd.NDArray or mxnet.sym.Symbol
            Dense array of center words of shape (batch_size, ). Only used for
            row-wise independently masking negatives equal to one of
            center_words.
        """

        # negatives sampling
        negatives = []
        mask = []
        for _ in range(self._kwargs['num_negatives']):
            negatives.append(self.negatives_sampler(center_words))
            mask_ = negatives[-1] != center_words
            mask_ = F.stack(mask_, (negatives[-1] != context))
            mask.append(mask_.min(axis=0))

        negatives = F.stack(*negatives, axis=1)
        mask = F.stack(*mask, axis=1).astype(np.float32)

        # center - context pairs
        emb_center = self.embedding(center).expand_dims(1)
        emb_context = self.embedding_out(context).expand_dims(2)
        pred_pos = F.batch_dot(emb_center, emb_context).squeeze()
        loss_pos = (F.relu(pred_pos) - pred_pos + F.Activation(
            -F.abs(pred_pos), act_type='softrelu')) / (mask.sum(axis=1) + 1)

        # center - negatives pairs
        emb_negatives = self.embedding_out(negatives).reshape(
            (-1, self._kwargs['num_negatives'],
             self._kwargs['output_dim'])).swapaxes(1, 2)
        pred_neg = F.batch_dot(emb_center, emb_negatives).squeeze()
        mask = mask.reshape((-1, self._kwargs['num_negatives']))
        loss_neg = (F.relu(pred_neg) + F.Activation(
            -F.abs(pred_neg), act_type='softrelu')) * mask
        loss_neg = loss_neg.sum(axis=1) / (mask.sum(axis=1) + 1)

        return loss_pos + loss_neg


class CBOW(Net):
    """CBOW network"""

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, center, context):
        """CBOW forward pass.

        Parameters
        ----------
        center : mxnet.nd.NDArray or mxnet.sym.Symbol
            Dense array of center words of shape (batch_size, ).
        context : mxnet.nd.NDArray or mxnet.sym.Symbol
            Sparse CSR array of word / subword indices of shape (batch_size,
            len(vocab) + num_subwords). Embedding for context words are
            computed via F.sparse.dot between the CSR center array and the
            weight matrix.

        """
        # negatives sampling
        negatives = []
        mask = []
        for _ in range(self._kwargs['num_negatives']):
            negatives.append(self.negatives_sampler(center))
            mask.append(negatives[-1] != center)

        negatives = F.stack(*negatives, axis=1)
        mask = F.stack(*mask, axis=1).astype(np.float32)

        # context - center samples
        emb_context = self.embedding(context).expand_dims(1)
        emb_center = self.embedding_out(center).expand_dims(2)
        pred_pos = F.batch_dot(emb_context, emb_center).squeeze()
        loss_pos = (F.relu(pred_pos) - pred_pos + F.Activation(
            -F.abs(pred_pos), act_type='softrelu')) / (mask.sum(axis=1) + 1)

        # context - negatives samples
        emb_negatives = self.embedding_out(negatives).reshape(
            (-1, self._kwargs['num_negatives'],
             self._kwargs['output_dim'])).swapaxes(1, 2)
        pred_neg = F.batch_dot(emb_context, emb_negatives).squeeze()
        mask = mask.reshape((-1, self._kwargs['num_negatives']))
        loss_neg = (F.relu(pred_neg) + F.Activation(
            -F.abs(pred_neg), act_type='softrelu')) * mask
        loss_neg = loss_neg.sum(axis=1) / (mask.sum(axis=1) + 1)

        return loss_pos + loss_neg


# Helper functions - train and eval

def norm_vecs_by_row(x):
    return x / (mx.nd.sum(x * x, axis=1) + 1e-10).sqrt().reshape((-1, 1))


def get_k_closest_offers(vocab, embedding, k, offer, offer_to_name):
    offer_vec = norm_vecs_by_row(embedding[[offer]])
    vocab_vecs = norm_vecs_by_row(embedding[vocab.idx_to_token])
    dot_prod = mx.nd.dot(vocab_vecs, offer_vec.T)
    indices = mx.nd.topk(
        dot_prod.reshape((len(vocab.idx_to_token), )),
        k=k+1,
        ret_typ='indices')
    indices = [int(i.asscalar()) for i in indices]
    result = [offer_to_name[vocab.idx_to_token[i]] for i in indices[1:]]
    logger.info('closest tokens to "%s":\n\t%s' % (offer_to_name[offer], "\n\t".join(result)))


def train_embedding(vocab, batches, embedding, trainer, ctx, num_epochs, log_interval, offer_to_name, model_save_path):
    for epoch in range(1, num_epochs + 1):
        start_time = time.time()
        l_avg = 0
        log_wc = 0

        logger.info('Beginning epoch %d and resampling data.' % epoch)
        for i, batch in enumerate(batches):
            batch = [array.as_in_context(ctx) for array in batch]
            with mx.autograd.record():
                loss = embedding(*batch)
            loss.backward()
            trainer.step(1)

            l_avg += loss.mean()
            log_wc += loss.shape[0]
            if i % log_interval == 0:
                # mx.nd.waitall()
                wps = log_wc / (time.time() - start_time)
                l_avg = l_avg.asscalar() / log_interval
                logger.info('epoch %d, iteration %d, loss %.2f, throughput=%.2fK wps' % (epoch, i, l_avg, wps / 1000))
                start_time = time.time()
                log_wc = 0
                l_avg = 0

        embedding.export(model_save_path, epoch + 1)
        logger.info("model has been saved")
        get_k_closest_offers(vocab, embedding, 10, 46, offer_to_name)  # 46: $8.99 Whopper Meal for Two
    return None


# Main functions

def get_offer2vec_features(offer_pool_all, offer_bag_data, model_save_path, ctx, mode, embedding_size=100,
                           batch_size=4096, window_size=3, frequent_token_subsampling=1E-1, num_negatives=5,
                           learning_rate=0.05, num_epochs=1, log_interval=500, **kwargs):
    assert mode in ["cbow", "CBOW", "sg", "SG"], "mode should be either 'CBOW' or 'SG'"
    data_raw = mx.gluon.data.dataset.SimpleDataset(offer_bag_data)
    offer_count = len(offer_pool_all)
    offer_to_name = {
        k: v["offer_name"]
        for k, v in offer_pool_all[["offer_name", "offer_idx"]].set_index("offer_idx").to_dict(orient="index").items()
    }
    is_cbow = True if mode in ["cbow", "CBOW"] else False

    # before indexing
    log_string = f"# sentences (user count): {len(data_raw)}\n"
    for i, sentence in enumerate(data_raw):
        log_string += f"\t# tokens: {len(sentence)} {sentence[:10]}\n"
        if i > 2:
            break
    logger.info(log_string)

    # build vocab and index
    data_idx, vocab, idx_to_counts = preprocess_dataset(
        data=data_raw,
        vocab_len=offer_count,
        min_freq=1,
        max_vocab_size=None
    )
    assert all([k == v for k, v in vocab.token_to_idx.items()]), "offer_idx != idx"

    # after indexing
    log_string = f"# sentences (user count): {len(data_raw)}\n"
    for sentences in data_idx:
        for i, sentence in enumerate(sentences):
            log_string += f"\t# tokens: {len(sentence)} {sentence[:10]}\n"
            if i > 2:
                break
    logger.info(log_string)

    # subsample and generate context-center pairs
    data, batchify_fn = transform_data_word2vec(
        data=data_idx,
        vocab=vocab,
        idx_to_counts=idx_to_counts,
        cbow=is_cbow,
        batch_size=batch_size,
        window_size=window_size,
        frequent_token_subsampling=frequent_token_subsampling,
    )

    # for i, d in enumerate(data):
    #     print(d)
    #     break

    # batchify
    batches = data.transform(batchify_fn)

    # for i, d in enumerate(batches):
    #     print(d)
    #     break

    negatives_weights = mx.nd.array(idx_to_counts)
    if is_cbow:
        embedding = CBOW(
            token_to_idx=vocab.token_to_idx,
            output_dim=embedding_size,
            negatives_weights=negatives_weights,
            subword_function=None,
            num_negatives=num_negatives,
            smoothing=0.75,
            sparse_grad=True,
            dtype="float32",
        )
    else:
        embedding = SG(
            token_to_idx=vocab.token_to_idx,
            output_dim=embedding_size,
            negatives_weights=negatives_weights,
            subword_function=None,
            num_negatives=num_negatives,
            smoothing=0.75,
            sparse_grad=True,
            dtype="float32",
        )
    embedding.initialize(ctx=ctx)
    embedding.hybridize()
    trainer = mx.gluon.Trainer(embedding.collect_params(), 'adagrad', dict(learning_rate=learning_rate))
    logger.info(embedding)

    train_embedding(
        vocab=vocab,
        batches=batches,
        embedding=embedding,
        trainer=trainer,
        ctx=ctx,
        num_epochs=num_epochs,
        log_interval=log_interval,
        offer_to_name=offer_to_name,
        model_save_path=model_save_path,
    )
    offer2vec_features = embedding[vocab.idx_to_token].asnumpy()

    return offer2vec_features




import numpy as np
import torch


def get_loss_scales(net, task1_criterion, task2_criterion, dataloader, device):
    net.eval()
    net.to(device)
    loss1_all = []
    loss2_all = []
    with torch.no_grad():
        for (ctx, seq, vl, candidate), (label_pos, label_price) in dataloader:
            out = net(
                ctx_in=[c.to(device) for c in ctx],
                seq_in=seq.to(device),
                vl_in=vl.to(device),
                candidate_in=candidate.to(device),
                seq_history=None
            )
            # print("*" * 10, f"torch.cuda.max_memory_allocated(): {torch.cuda.max_memory_allocated()}, "
            #       f"memory_allocated: {torch.cuda.memory_allocated()}", "*" * 10)
            loss1 = task1_criterion(out[0].squeeze(), label_pos.float().to(device))
            loss2 = task2_criterion(out[1].squeeze(), label_price.float().to(device))
            loss1_all.append(loss1.item())
            loss2_all.append(loss2.item())
    print("mean:", np.mean(loss1_all), np.mean(loss2_all))
    print("max:", np.max(loss1_all), np.max(loss2_all))
    loss1_scale = np.max(loss1_all)
    loss2_scale = np.max(loss2_all)
    return loss1_scale, loss2_scale


class BaseLogger:
    def __init__(self, metric):
        assert metric in ["acc", "f1", "mae", "mse", "loss", "identity"]
        self.metric_name = metric
        if metric == "acc":
            self.metric_fn = self.get_acc
        elif metric == "f1":
            self.metric_fn = self.get_f1
        elif metric == "mae":
            self.metric_fn = self.get_mae
        elif metric == "mse":
            self.metric_fn = self.get_mse
        elif metric == "identity" or metric == "loss":
            self.metric_fn = self.get_loss
        else:
            raise NotImplementedError()
        self.metrics_step = []

    def update(self, y_true, y_pred):
        self.metrics.append(self.metric_fn(y_true, y_pred))

    def get(self, reduce="mean"):
        if reduce == "mean":
            return np.mean(self.metrics)
        elif reduce == "count":
            return len(self.metrics)
        else:
            raise NotImplementedError()

    def reset(self):
        self.metrics = []

    @staticmethod
    def get_loss(loss, _):
        return loss.item()

    @staticmethod
    def get_acc(y_true: torch.Tensor, y_pred: torch.Tensor):
        return (y_true == y_pred).float().mean().item()

    @staticmethod
    def get_f1(y_true, y_pred, epsilon=1e-7):
        tp = (y_true * y_pred).sum().float()
        # tn = ((1 - y_true) * (1 - y_pred)).sum().float()
        fp = ((1 - y_true) * y_pred).sum().float()
        fn = (y_true * (1 - y_pred)).sum().float()
        precision = tp / (tp + fp + epsilon)
        recall = tp / (tp + fn + epsilon)
        f1 = 2 * (precision * recall) / (precision + recall + epsilon)
        return f1.item()

    @staticmethod
    def get_mae(y_true: torch.Tensor, y_pred: torch.Tensor):
        return (y_true - y_pred).abs().mean().item()

    @staticmethod
    def get_mse(y_true: torch.Tensor, y_pred: torch.Tensor):
        return (y_true - y_pred).pow(2).mean().item()


class MetricLogger:
    def __init__(self, names, metrics):
        self.metrics = {
            name: BaseLogger(metric=metric)
            for name, metric in zip(names, metrics)
        }

    def add(self, name, metric):
        self.metrics.update({name: metric})

    def update(self, metric_name, y_true, y_pred):
        if metric_name is None:
            for metric in self.metrics.values():
                metric.update(y_true, y_pred)
        else:
            self.metrics[metric_name].update(y_true, y_pred)

    def reset(self, metric_name):
        if metric_name is None:
            for metric in self.metrics.values():
                metric.reset()
        else:
            self.metrics[metric_name].reset()

    def get(self, metric_name, reduce="mean"):
        return self.metrics[metric_name].get(reduce=reduce)
